/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *  Hosts.java
 *  Copyright (C) 2007 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */

package weka4ws.core;

import java.awt.BorderLayout;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.StringTokenizer;

import javax.swing.BorderFactory;
import javax.swing.DefaultComboBoxModel;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;

import org.globus.gsi.GlobusCredentialException;

/**
 * Sets up and stores the remote location addresses where the computational
 * tasks will be executed. Currently they are loaded from a file resident on the
 * local machine: the possibility to get them from the "Discovery Service"
 * provided by the Globus Toolokit is still under development.
 * <p>
 * The first two elements of the arrays <code>addresses</code>,
 * <code>servicePorts</code> and <code>gridFTPPorts</code> are reserved for
 * the local and automatic options respectively.
 * 
 * @author Marco Lackovic
 * @version 1.0
 * @see weka4ws.core.WUtil
 * @see weka4ws.core.HostCheckConnectionThread
 * @see weka4ws.core.HostCheckThread
 */
public class Hosts {

    /**
     * Used in a ComboBox in the Classifier, Clusterer and Association panels,
     * inside the Explorer component,
     */
    public static DefaultComboBoxModel CBModel = new DefaultComboBoxModel();

    protected static JTextArea logBoard;

    protected static JFrame logFrame;

    public static final String LOCAL = "localhost";

    public static final String AUTO = "auto";

    public static final short LOCAL_INDEX = 0;

    public static final short AUTO_INDEX = 1;

    private static String[] remoteHosts;

    public static String[] servicePorts;

    public static String[] gridFTPPorts;

    public static String[] loggingOptions;

    private static boolean first = true;

    public static boolean loading = false;

    protected static int numTotal;

    protected static int workingHosts;

    private static int lastChosenHost = 0;

    private static final String machinesFilePath = "etc/machines";

    public static String getHostAt(int i) {
        return remoteHosts[i];
    }

    public static String[] getHostsArray() {
        return remoteHosts;
    }

    protected static void logMessage(String message) {

        if (first) {
            first = false;
        } else {
            logBoard.append("\n");
        }
        logBoard.append(message);
        logBoard.setCaretPosition(logBoard.getText().length());
    }

    protected static int getHostsNumberOnFile() {

        int num = 0;
        String line = new String();
        BufferedReader out;

        try {
            out = new BufferedReader(new FileReader(machinesFilePath));
            while ((line = out.readLine()) != null) {
                if (!line.startsWith("#") && (line.trim().length() != 0)) {
                    num++;
                }
            }
            out.close();
        } catch (FileNotFoundException e) {
            WUtil.showDialog(null, e.getMessage(), "File not found",
                    JOptionPane.ERROR_MESSAGE);
        } catch (IOException e) {
            e.printStackTrace();
        }
        return num;
    }

    public static void initialization() {
        logBoard = new JTextArea(4, 40);
        logFrame = new JFrame("Hosts log");
        logBoard.setEditable(false);
        logBoard.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
        final JScrollPane js = new JScrollPane(logBoard);

        logFrame.addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                logFrame.setVisible(false);
            }
        });
        logFrame.getContentPane().setLayout(new BorderLayout());
        logFrame.getContentPane().add(js, BorderLayout.CENTER);
        logFrame.pack();
        logFrame.setSize(500, 350);
        logFrame.setLocation(300, 0);
    }

    public synchronized static void load(boolean show) {

        loading = true;

        // getMachinesInfo: reads information from "machines" file
        ArrayList remHostsAList = new ArrayList();
        ArrayList remServicePortsAList = new ArrayList();
        ArrayList remGridFTPPortsAList = new ArrayList();
        ArrayList loggingOptionsAList = new ArrayList();

        remHostsAList.add(LOCAL);
        remServicePortsAList.add("0");
        remGridFTPPortsAList.add("0");
        loggingOptionsAList.add("0");

        remHostsAList.add(AUTO);
        remServicePortsAList.add("0");
        remGridFTPPortsAList.add("0");
        loggingOptionsAList.add("0");

        CBModel.removeAllElements();
        CBModel.addElement(LOCAL);
        CBModel.addElement(AUTO);

        try {
            String line = new String();
            BufferedReader out = new BufferedReader(new FileReader(
                    machinesFilePath));

            logFrame.setVisible(show);
            numTotal = getHostsNumberOnFile();
            logMessage("Adding " + numTotal + " hosts...");
            int i = 1;
            while ((line = out.readLine()) != null) {
                if (!line.startsWith("#") && (line.trim().length() != 0)) {
                    StringTokenizer lineTokenizer = new StringTokenizer(line);
                    String remHost = lineTokenizer.nextToken();
                    String remServicePort = lineTokenizer.nextToken();
                    String remGridFTPPort = lineTokenizer.nextToken();
                    String logging = lineTokenizer.nextToken();

                    remHostsAList.add(remHost);
                    remServicePortsAList.add(remServicePort);
                    remGridFTPPortsAList.add(remGridFTPPort);
                    loggingOptionsAList.add(logging);

                    logMessage(i++ + ". " + remHost + " added successfully.");
                    workingHosts++;
                    CBModel.addElement(remHost);
                }
            }
            workingHosts = 0;
            logMessage("Hosts checking completed, " + workingHosts
                    + " added successfully.");
            logMessage("-------------------------");
            out.close();
        } catch (IOException e) {
            e.printStackTrace();
        }

        int size = remHostsAList.size();
        remoteHosts = new String[size];
        servicePorts = new String[size];
        gridFTPPorts = new String[size];
        loggingOptions = new String[size];

        remHostsAList.toArray(remoteHosts);
        remServicePortsAList.toArray(servicePorts);
        remGridFTPPortsAList.toArray(gridFTPPorts);
        loggingOptionsAList.toArray(loggingOptions);

        loading = false;
    }

    public synchronized static void loadAndCheck(boolean show) {

        loading = true;

        // getMachinesInfo: reads information from "machines" file
        ArrayList remHostsAList = new ArrayList();
        ArrayList remServicePortsAList = new ArrayList();
        ArrayList remGridFTPPortsAList = new ArrayList();
        ArrayList loggingOptionsAList = new ArrayList();

        remHostsAList.add(LOCAL);
        remServicePortsAList.add("0");
        remGridFTPPortsAList.add("0");
        loggingOptionsAList.add("0");

        remHostsAList.add(AUTO);
        remServicePortsAList.add("0");
        remGridFTPPortsAList.add("0");
        loggingOptionsAList.add("0");

        CBModel.removeAllElements();
        CBModel.addElement(LOCAL);
        CBModel.addElement(AUTO);

        try {
            String line = new String();
            BufferedReader out = new BufferedReader(new FileReader(
                    machinesFilePath));

            logFrame.setVisible(show);
            logMessage("    Verifying proxy credentials...");
            if (WUtil.proxyTimeLeft() == 0) {
                logMessage("    Proxy credentials have expired: generating a new proxy...");
                WUtil.visualGridProxyInit();
            }
            WUtil.getCredential().verify();
            logMessage("    Valid proxy credentials detected.");
            numTotal = getHostsNumberOnFile();
            logMessage("    "
                    + numTotal
                    + " hosts found on \"etc/machines\". Beginning verification.");
            ArrayList remHosts = new ArrayList();
            while ((line = out.readLine()) != null) {
                if (!line.startsWith("#") && (line.trim().length() != 0)) {
                    StringTokenizer lineTokenizer = new StringTokenizer(line);
                    String remAddress = lineTokenizer.nextToken();
                    String remServicePort = lineTokenizer.nextToken();
                    String remGridFTPPort = lineTokenizer.nextToken();
                    String logging = lineTokenizer.nextToken();

                    HostCheckThread remHost = new HostCheckThread(remAddress,
                            remServicePort, remGridFTPPort, logging);
                    remHost.start();
                    remHosts.add(remHost);
                }
            }
            HostCheckThread remHost;
            workingHosts = 0;
            for (int i = 0; i < numTotal; i++) {
                remHost = (HostCheckThread) remHosts.get(i);
                try {
                    remHost.join();
                    String host = remHost.getHost();
                    if (remHost.isFunctional()) {
                        remHostsAList.add(host);
                        remServicePortsAList.add(remHost.getServicePort());
                        remGridFTPPortsAList.add(remHost.getGridFTPPort());
                        loggingOptionsAList.add(remHost.getLogging());
                        logMessage("+ " + (i + 1) + ". Host \"" + host
                                + "\" added successfully.");
                        workingHosts++;
                        CBModel.addElement(host);
                    } else {
                        logMessage("   " + (i + 1) + ". "
                                + remHost.getMessage());
                    }
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }

            }
            logMessage("    Verification completed: " + workingHosts
                    + " operative hosts found");
            logMessage("------------ " + WUtil.getDateTime() + " -------------");
            out.close();
        } catch (IOException e) {
            WUtil.showDialog(null, e.getMessage(), "File not found",
                    JOptionPane.ERROR_MESSAGE);
        } catch (GlobusCredentialException e) {
            logMessage("    " + e.getMessage());
            logMessage("    Error: can't check hosts locations until a new proxy is generated.");
        }

        int size = remHostsAList.size();
        remoteHosts = new String[size];
        servicePorts = new String[size];
        gridFTPPorts = new String[size];
        loggingOptions = new String[size];

        remHostsAList.toArray(remoteHosts);
        remServicePortsAList.toArray(servicePorts);
        remGridFTPPortsAList.toArray(gridFTPPorts);
        loggingOptionsAList.toArray(loggingOptions);

        loading = false;
    }

    public synchronized static String[] getAutoHost()
            throws NoHostsFoundException {
        if (workingHosts == 0) {
            throw new NoHostsFoundException();
        }
        String[] parameters = new String[4];
        int chosen;
        if (numTotal == 1)
            chosen = 2;
        else {
            chosen = lastChosenHost + 2;
            lastChosenHost = (lastChosenHost + 1) % workingHosts;
        }
        parameters[0] = remoteHosts[chosen];
        parameters[1] = servicePorts[chosen];
        parameters[2] = gridFTPPorts[chosen];
        parameters[3] = loggingOptions[chosen];
        return parameters;

    }

    public static DefaultComboBoxModel getNewCBModel() {
        return new DefaultComboBoxModel(remoteHosts);

    }

    public static int getNumTotal() {
        return numTotal;
    }

}
