/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *  RemoteMining.java
 *  Copyright (C) 2007 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */

package weka4ws.core;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.net.URL;
import java.rmi.RemoteException;
import java.util.Random;
import java.util.zip.CRC32;
import java.util.zip.Checksum;

import javax.xml.rpc.ServiceException;

import org.apache.axis.types.URI.MalformedURIException;
import org.globus.wsrf.NotificationConsumerManager;

import weka.core.Utils;
import weka.gui.Logger;
import weka4ws.stubs.service.Algorithm;
import weka4ws.stubs.service.Dataset;
import weka4ws.stubs.service.Response;
import weka4ws.stubs.service.Result;
import weka4ws.stubs.service.StopParameters;
import weka4ws.stubs.service.Weka4WSException;
import weka4ws.stubs.service.Weka4WSPortType;

/**
 * Handles all the transactions with the service node.
 * 
 * @author <a href="mailto:mlackovic@deis.unical.it">Marco Lackovic</a>
 */
public abstract class RemoteTask implements RemoteTaskConstants {

    protected String hostAddress = "localhost";

    protected int gridFTPPort = 2811;

    protected int servicePort = 8443;

    protected boolean logging = false;

    protected Dataset dataset = new Dataset();

    protected Dataset testset = new Dataset();

    protected Logger log;

    protected RemoteParameters parameters = null;

    protected long time;

    protected long trainTimeStart;

    protected String executionTime;

    protected long taskID;

    protected Weka4WSPortType port;

    protected RemoteResourse resource;

    private NotificationReceiver receiver;

    private static TransfersManager tManager = new TransfersManager();

    public RemoteTask(ConnectionParameters connectionParameters,
            RemoteParameters parameters) throws MalformedURIException,
            RemoteException, ServiceException {

        trainTimeStart = System.currentTimeMillis();
        time = trainTimeStart;

        hostAddress = connectionParameters.getRemoteHost();
        gridFTPPort = connectionParameters.getRemoteGridFTPPort();
        servicePort = connectionParameters.getRemoteServicePort();
        logging = connectionParameters.getRemoteLogging();

        log = connectionParameters.getLog();
        log
                .statusMessage("Weka4WS: Remote datamining in progress. See log for details.");
        logMessage("=> Setting up remote task");

        this.parameters = parameters;

        Random randomGenerator = new Random();
        taskID = Math.abs(randomGenerator.nextLong());
        this.parameters.setTaskID(taskID);

        this.parameters.setLogging(logging);

        logMessage("   Remote resource creation");
        resource = new RemoteResourse(hostAddress, servicePort, "Result");
    }

    protected void logMessage(String s) {
        long now = System.currentTimeMillis();
        WUtil.logMessage(log, hostAddress, s + " ("
                + Utils.doubleToString((now - time) / 1000.0, 2) + "s)");
        time = now;
    }

    private Response invokeService() throws RemoteException {

        /* ============== Weka4WS modifications ============== */
        // Output some header information
        logMessage("   Invoked " + parameters.getAlgorithm().getName());
        /* =================================================== */

        if (this instanceof RemoteClassifier)
            return port.classification((ClassificationParameters) parameters);
        if (this instanceof RemoteClusterer)
            return port.clustering((ClusteringParameters) parameters);
        if (this instanceof RemoteAssociator)
            return port
                    .associationRules((AssociationRulesParameters) parameters);
        return null;
    }

    /**
     * Returns a String containg the model built at the host
     * 
     * @return Object
     * @throws Exception
     */
    public Result getResult() throws Exception {

        if (this.parameters == null) {
            throw new Exception("Parameters not initialized");
        }

        port = resource.getPortType();

        receiver = new NotificationReceiver(hostAddress, log);
        NotificationConsumerManager consumer = NotificationConsumerManager
                .getInstance();
        logMessage("   Notification subscription (default style: pull mode)");
        resource.notifSubscribe(receiver, consumer);

        Transfer transfer = new Transfer(dataset.getFileName(), dataset
                .getCrc(), hostAddress);
        boolean lock = tManager.getLock(transfer);
        // First attempt to start the remote computation
        Response response = invokeService();
        if (response.getException().isThrown()) {
            throw new Exception(response.getException().getMessage());
        }
        if (lock) {
            boolean transferPerformed = checkRemoteFile(response);

            if (transferPerformed) {
                // Second and last attempt to start the remote computation
                invokeService();
                if (response.getException().isThrown()) {
                    throw new Exception(response.getException().getMessage());
                }
            }

            tManager.releaseLock(transfer);
        }

        logMessage("   Remote data mining...");
        Result result = receiver.getResult(port);
        if (result != null) {
            logMessage("<= Results received.");

            long trainTimeElapsed = System.currentTimeMillis() - trainTimeStart;
            executionTime = Utils.doubleToString(trainTimeElapsed / 1000.0, 2);
            logMessage("   Time taken to build "
                    + parameters.getAlgorithm().getName().substring(
                            parameters.getAlgorithm().getName()
                                    .lastIndexOf('.') + 1) + " model: "
                    + WUtil.showTime(trainTimeElapsed));

            logMessage("   Resource destruction");
            resource.destroy();

            if (consumer != null) {
                try {
                    consumer.stopListening();
                } catch (Exception ee) {
                }
            }

            Weka4WSException remoteException = result.getException();
            if (remoteException.isThrown()) {
                throw new Exception(remoteException.getMessage());
            }
        }
        return result;

    }

    /**
     * Checks whether a file is available on a remote host. If not already
     * available, the file is copied to the remote host.
     */
    public boolean checkRemoteFile(Response response) {
        boolean transferPerformed = false;

        try {
            if (!response.isDatasetFound()) {
                logMessage("   Transferring dataset file \""
                        + dataset.getFileName() + "\"");
                WUtil.performTransfer(dataset, hostAddress, gridFTPPort,
                        response.getDirPath());
                transferPerformed = true;
                logMessage("   Dataset file \"" + dataset.getFileName()
                        + "\" transferred successfully");
            } else {
                logMessage("   Transfer not performed, dataset \""
                        + dataset.getFileName() + "\" already available");
            }
            if (((this instanceof RemoteClassifier) && (((ClassificationParameters) parameters)
                    .getTestOptions().getTestMode() == TEST_SET))
                    || ((this instanceof RemoteClusterer) && (((ClusteringParameters) parameters)
                            .getTestOptions().getTestMode() == TEST_SET))) {
                if (!response.isTestsetFound()) {
                    logMessage("   Transferring testset file \""
                            + testset.getFileName() + "\"");
                    WUtil.performTransfer(testset, hostAddress, gridFTPPort,
                            response.getDirPath());
                    transferPerformed = true;
                    logMessage("   Testset file \"" + testset.getFileName()
                            + "\" transferred successfully");
                } else {
                    logMessage("   Transfer not performed, testset \""
                            + testset.getFileName() + "\" already available");
                }
            }

        } catch (Exception e) {
            e.printStackTrace();
        }

        return transferPerformed;
    }

    public void setDataset(String datasetFilePath) throws Exception {

        String filePath;
        if (datasetFilePath.startsWith("http://")) {
            URL url = new URL(datasetFilePath);
            String fileName = File.separator
                    + (new File(url.getFile())).getName();
            String currentDir = System.getProperty("user.home");
            // TODO Aggiungere una throw exception se la creazione della
            // cartella non va a buon fine
            (new File(currentDir + File.separator + TEMP_DIR)).mkdir();
            WUtil.downloadDataset(datasetFilePath, fileName, TEMP_DIR);
            filePath = currentDir + File.separator + TEMP_DIR + File.separator
                    + fileName;
        } else {
            filePath = datasetFilePath;
        }

        long crcValue = getChecksumValue(filePath);

        dataset.setCrc(crcValue);

        dataset.setFileName(filePath.substring(filePath
                .lastIndexOf(File.separator) + 1));

        dataset.setFilePath(filePath);

        this.parameters.setDataset(dataset);
    }

    public void setAlgorithm(String name, String parameters) {
        Algorithm alg = new Algorithm();
        alg.setName(name);
        alg.setParameters(parameters);
        this.parameters.setAlgorithm(alg);
    }

    /**
     * Computes the CRC.
     */
    protected long getChecksumValue(String filePath) {
        Checksum checksum = new CRC32();
        try {
            BufferedInputStream is = new BufferedInputStream(
                    new FileInputStream(filePath));
            byte[] bytes = new byte[1024];
            int len = 0;
            while ((len = is.read(bytes)) >= 0) {
                checksum.update(bytes, 0, len);
            }
            is.close();
        } catch (IOException e) {
            e.printStackTrace();
        }
        return checksum.getValue();
    }

    /**
     * @return the executionTime
     * @uml.property name="executionTime"
     */
    public String getExecutionTime() {
        return executionTime;
    }

    public String getLocation() {
        return hostAddress;
    }

    public void stop() throws RemoteException {
        if (port != null) {
            StopParameters stopParameters = new StopParameters();
            stopParameters.setTaskID(taskID);
            port.stopTask(stopParameters);
            receiver.stop();
            resource.destroy();
        }
    }

}
