/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *  HostCheckConnectionThread.java
 *  Copyright (C) 2005 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */

package weka4ws.core;

import java.io.BufferedInputStream;
import java.io.BufferedWriter;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Writer;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Locale;
import java.util.zip.GZIPOutputStream;

import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;

import org.apache.axis.encoding.Base64;
import org.globus.axis.util.Util;
import org.globus.ftp.DataChannelAuthentication;
import org.globus.ftp.GridFTPClient;
import org.globus.ftp.Session;
import org.globus.gsi.GlobusCredential;
import org.globus.gsi.GlobusCredentialException;
import org.globus.io.urlcopy.UrlCopy;
import org.globus.tools.proxy.GridProxyInit;
import org.globus.util.GlobusURL;

import weka.core.Instances;
import weka.gui.Logger;
import weka4ws.stubs.service.Dataset;

/** 
 * A collection of static methods and fields useful for Weka4WS Explorer and
 * Knowledge Flow component.
 *
 * @author Marco Lackovic
 * @author Oreste Verta
 * @version $Revision: 1.0 $
 */
public class WUtil {

    public static final String VERSION = "2.1";

    private static boolean initialized = false;

    private static FileManager fManager = new FileManager();

    public static Thread loadThread = new Thread() {
        public synchronized void run() {
            Hosts.loadAndCheck(true);
        }
    };

    private static String GLOBUS_LOCATION;

    public static void initialization() {
        if (!initialized) {
            checkGlobusLocation();
            registerTransport();
            Hosts.initialization();
            loadThread.start();
            initialized = true;
        }
    }

    /**
     * Registers the Axis transport handler for "https" (for transport security),
     * only if it hasn't been already registered.
     */
    public static void registerTransport() {
        Util.registerTransport();
    }

    /**
     * Checks whether the environment variable GLOBUS_LOCATION is properly
     * set; if not, print an error in the console and terminates the
     * application.
     */
    public static void checkGlobusLocation() {
        GLOBUS_LOCATION = System.getProperty("GLOBUS_LOCATION");
        if (GLOBUS_LOCATION == null) {
            System.err
                    .println("Environment variable GLOBUS_LOCATION is not defined. Try starting the JVM with the \"-DGLOBUS_LOCATION=$GLOBUS_LOCATION\" switch.");
            System.exit(0);
        }
    }

    public static void logMessage(Logger log, String hostAddress, String s) {
        synchronized (log) {
            log.logMessage(hostAddress + ": " + s);
        }
    }

    public static void copyFile(String inFileName, String outFileName)
            throws Exception {
        InputStream is = ClassLoader.getSystemResourceAsStream(inFileName);

        File tempFile = new File(outFileName);
        tempFile.deleteOnExit();

        FileOutputStream fos = new FileOutputStream(tempFile);

        byte[] buf = new byte[1024];
        int i = 0;
        while ((i = is.read(buf)) != -1) {
            fos.write(buf, 0, i);
        }
        is.close();
        fos.close();
    }

    public static void copyStreamToFile(String sourceFilePath,
            String destinationFilePath) throws Exception {

        InputStream inFileStream = ClassLoader
                .getSystemResourceAsStream(sourceFilePath);

        BufferedInputStream buf = new BufferedInputStream(inFileStream);
        FileOutputStream fout = null;

        fout = new FileOutputStream(destinationFilePath);

        byte[] buffer = new byte[1024];//byte buffer
        int bytesRead = 0;
        while (true) {
            bytesRead = buf.read(buffer, 0, 1024);
            //bytesRead returns the actual number of bytes read from
            //the stream. returns -1 when end of stream is detected
            if (bytesRead == -1)
                break;
            fout.write(buffer, 0, bytesRead);
        }
        buf.close();
        fout.close();
        buf = null;
        fout = null;
    }

    public static String showTime(long milliSeconds) {

        long seconds = milliSeconds / 1000;
        long hours = seconds / 3600;
        long minutes = seconds / 60 - hours * 60;
        long sec = seconds - (hours * 3600 + minutes * 60);

        String outputString;

        if (hours != 0) {
            outputString = "[ " + hours + "h " + minutes + "m " + sec + "s ("
                    + seconds + "s) ]";
        } else if (minutes != 0) {
            outputString = "[ " + minutes + "m " + sec + "s (" + seconds
                    + "s) ]";
        } else {
            outputString = "[ " + sec + "s ]";
        }

        return outputString;
    }

    /**
     * Saves the instances in a file, located in the user's hidden subfolder
     * named ".Weka4WS", having the name of the original dataset plus 
     * (for uniquity) the number of milliseconds since January 1, 1970.
     */
    public static String saveWorkingInstancesToFile(String fileName,
            Instances instances) {

        fileName += new Date().getTime() + Instances.FILE_EXTENSION;
        String saveDir = System.getProperty("user.home") + File.separator
                + RemoteTaskConstants.TEMP_DIR;
        if (!(new File(saveDir)).exists()) {
            (new File(saveDir)).mkdir();
        }
        String newDatasetFilePath = saveDir + File.separator + fileName;
        File newDatasetFile = new File(newDatasetFilePath);
        saveInstancesToFile(newDatasetFile, instances);
        return newDatasetFilePath;
    }

    /**
     * Saves the current instances to the supplied file.
     *
     * @param f a value of type 'File'
     * @param inst the instances to save
     * @param saveHeader true to save in arff format, false to save in csv
     */
    private static void saveInstancesToFile(final File f, final Instances inst) {

        try {
            Writer w = new BufferedWriter(new FileWriter(f));

            Instances h = new Instances(inst, 0);
            w.write(h.toString());
            w.write("\n");

            for (int i = 0; i < inst.numInstances(); i++) {
                w.write(inst.instance(i).toString());
                w.write("\n");
            }
            w.close();
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }

    public static Object getObjFromStr(String modelString) {

        if ((modelString == null) || (modelString == "")
                || (modelString.length() == 0)) {
            return null;
        }

        Object model = null;
        ByteArrayInputStream byteStream = new ByteArrayInputStream(Base64
                .decode(modelString));
        ObjectInputStream inputStream = null;
        try {
            inputStream = new ObjectInputStream(byteStream);
            model = inputStream.readObject();
            inputStream.close();
        } catch (IOException ex) {
            ex.printStackTrace();
        } catch (ClassNotFoundException ex) {
            ex.printStackTrace();
        }

        return model;
    }

    public static String getStrFromObj(Object model) {
        if (model == null) {
            return "";
        }
        ByteArrayOutputStream byteStream = new ByteArrayOutputStream();
        ObjectOutputStream outputStream = null;
        String encodedString = null;

        try {
            outputStream = new ObjectOutputStream(byteStream);
            outputStream.writeObject(model);
            outputStream.close();
            encodedString = Base64.encode(byteStream.toByteArray());
        } catch (IOException ex) {
            ex.printStackTrace();
        }
        return encodedString;
    }

    static String makeGZip(String inFilePath) {
        String outFilePath = inFilePath + ".gz";
        try {

            GZIPOutputStream out = new GZIPOutputStream(new FileOutputStream(
                    outFilePath));

            // Open the input file
            FileInputStream in = new FileInputStream(inFilePath);

            // Transfer bytes from the input file to the GZIP output stream
            byte[] buf = new byte[1024];
            int len;
            while ((len = in.read(buf)) > 0) {
                out.write(buf, 0, len);
            }
            in.close();

            // Complete the GZIP file
            out.finish();
            out.close();

        } catch (IOException e) {
        }
        return outFilePath;
    }

    /**
     * Donwloads a file from a URL to the local filesystem.
     */
    public static void downloadDataset(String fileURL, String fileName,
            String tempDir) throws Exception {
        GlobusURL source = new GlobusURL(fileURL);
        GlobusURL destination = new GlobusURL("file://" + "./" + tempDir
                + fileName);
        UrlCopy copy = new UrlCopy();
        copy.setSourceUrl(source);
        copy.setDestinationUrl(destination);
        // copy.setUseThirdPartyCopy(true);
        copy.copy();
    }

    /**
     * Performs a file transfer using GridFTP at the server node.
     */
    public static void performTransfer(Dataset d, String remoteHost, int port,
            String remoteUploadDir) {
        try {
            DatasetFile df = new DatasetFile(d.getFileName(), d.getCrc());
            String compressedFilePath = d.getFilePath() + ".gz";
            if (fManager.allowedToCompress(df)) {
                makeGZip(d.getFilePath());
                fManager.fileCompressed(df);
            }

            File compressedFile = new File(compressedFilePath);
            String destinationFilePath = remoteUploadDir
                    + compressedFile.getName();

            GridFTPClient client = new GridFTPClient(remoteHost, port);
            client.authenticate(null);
            client.setDataChannelAuthentication(DataChannelAuthentication.NONE);
            client.setType(Session.TYPE_IMAGE);
            client.put(compressedFile, destinationFilePath, false);
            client.close();
            if (fManager.allowedToDelete(df)) {
                compressedFile.delete();
            }

        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    private static JOptionPane getNarrowOptionPane(int maxCharactersPerLineCount) {
        // Our inner class definition
        class NarrowOptionPane extends JOptionPane {
            private static final long serialVersionUID = 1L;
            int maxCharactersPerLineCount;

            NarrowOptionPane(int maxCharactersPerLineCount) {
                this.maxCharactersPerLineCount = maxCharactersPerLineCount;
            }

            public int getMaxCharactersPerLineCount() {
                return maxCharactersPerLineCount;
            }
        }

        return new NarrowOptionPane(maxCharactersPerLineCount);
    }

    public static void showDialog(JPanel panel, String message, String title,
            int errorMessage) {
        JOptionPane pane = getNarrowOptionPane(50);
        pane.setMessage(message);
        pane.setMessageType(errorMessage);
        pane.createDialog(panel, title).show();
    }

    /**
     * @return Seconds until proxy expires.
     * @throws GlobusCredentialException 
     */
    public static long proxyTimeLeft() {
        try {
            return getCredential().getTimeLeft();
        } catch (GlobusCredentialException e) {
            return 0;
        }
    }

    public static GlobusCredential getCredential()
            throws GlobusCredentialException {
        return GlobusCredential.getDefaultCredential();
    }

    public static void visualGridProxyInit() {
        final GridProxyInit proxy = new GridProxyInit();
        proxy.saveProxy(true);
        proxy.setCloseOnSuccess(true);
        proxy.pack();
        proxy.setLocation(400, 300);
        proxy.setModal(true);
        String sysLookAndFeel = UIManager.getSystemLookAndFeelClassName();
        try {
            UIManager.setLookAndFeel(sysLookAndFeel);
            SwingUtilities.updateComponentTreeUI(proxy);
        } catch (Exception e) {
            e.printStackTrace();
        }
        proxy.show();
    }

    public static String getDateTime() {
        DateFormat dateFormat = new SimpleDateFormat(
                "EEEEE, d MMMMM yyyy, HH:mm:ss", Locale.ENGLISH);
        Date date = new Date();
        return dateFormat.format(date);
    }
}
