/*
 *  ComputingNode.java
 *  Copyright (C) 2007 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */

package weka4ws.gui.kflow;

import javax.swing.JButton;
import javax.swing.JPanel;

import weka.gui.Logger;
import weka4ws.core.Hosts;
import weka4ws.core.NoHostsFoundException;
import weka4ws.core.WUtil;
import weka4ws.stubs.service.EvalResult;
import weka4ws.stubs.service.ModelResult;
import weka4ws.stubs.service.TestOptions;
import weka4ws.stubs.service.VisResult;

/**
 * Abstract class of something which can be remotely executed.
 * 
 * @author Marco Lackovic
 * @version 1.0
 * @see weka4ws.gui.kflow.Classifier
 * @see weka4ws.gui.kflow.Clusterer
 * @see weka4ws.gui.kflow.ClassifierPerformanceEvaluator
 * @see weka4ws.gui.kflow.TestSetMaker
 */
public abstract class ComputingNode extends JPanel {

    private static final long serialVersionUID = 1L;

    protected int hostSelectedIndex = Hosts.LOCAL_INDEX;

    protected String remoteHost = Hosts.LOCAL;

    protected int remoteServicePort = 0;

    protected int remoteGridFTPPort = 0;

    protected int remoteLogging = 0;

    protected TestSetNotifierThread connectedTSProducer;

    private boolean modelBuilt = false;

    protected static int IDLE = 0;

    protected static int BUILDING_MODEL = 1;

    protected static int CLUSTERING = 2;

    protected static int CLASSIFYING = 2;

    protected int m_state = IDLE;

    protected transient Logger m_log = null;

    protected JButton m_stopB;

    protected JButton m_startAllB;

    protected String dataSetFilePath;

    protected String cname;

    protected TestOptions testOptions;

    protected int testMode;

    protected int rndSeed;

    protected transient ModelResult modelResult;

    protected transient EvalResult evalResult;

    protected transient VisResult visResult;

    static protected int runningNodes = 0;

    protected boolean lastOne = false;

    static private long trainTotalTimeStart = 0;

    protected synchronized void increaseCounter() {
        if (runningNodes == 0) {
            trainTotalTimeStart = System.currentTimeMillis();
        }
        lastOne = false;
        runningNodes++;
    }

    protected synchronized void decreaseCounter() {
        runningNodes--;
        if (runningNodes == 0) {
            lastOne = true;
        }
    }

    protected synchronized boolean isLastOne() {
        return lastOne;
    }

    protected EvalResult getEval() {
        return evalResult;
    }

    protected VisResult getVis() {
        return visResult;
    }

    /**
     * Sets up the remote location address where the computational task will be
     * executed.
     */
    protected void setCompLocation(int hostSelected) {
        this.hostSelectedIndex = hostSelected;
        this.remoteHost = Hosts.getHostAt(hostSelectedIndex);
    }

    /**
     * Returns the remote location address where the computational task will be
     * executed.
     */
    protected int getSelectedHostIndex() {
        if ((hostSelectedIndex >= Hosts.getNumTotal())
                || !remoteHost.equals(Hosts.getHostAt(hostSelectedIndex))) {
            this.remoteHost = Hosts.LOCAL;
            this.hostSelectedIndex = 0;
        }
        return this.hostSelectedIndex;
    }

    /**
     * Tells the computing node there is a TestSetMaker which is waiting for its
     * completion.
     */
    protected synchronized void registerForNotification(
            TestSetNotifierThread tsnt) {
        connectedTSProducer = tsnt;
    }

    protected synchronized void unregister(TestSetNotifierThread tsnt) {
        if (connectedTSProducer == tsnt)
            connectedTSProducer = null;
    }

    protected synchronized boolean isModelBuilt() {
        return modelBuilt;
    }

    protected synchronized void setModelBuilt(boolean value) {
        modelBuilt = value;
    }

    protected boolean isRunning() {
        return (m_state != IDLE);
    }

    protected void logMessage(String s) {
        WUtil.logMessage(m_log, remoteHost, s);
    }

    protected synchronized void initParameters() {
        setModelBuilt(false);

        String host = Hosts.getHostAt(hostSelectedIndex);
        if (host.equals(Hosts.AUTO)) {
            try {
                String[] parameters = new String[3];
                parameters = Hosts.getAutoHost();
                remoteHost = parameters[0];
                remoteServicePort = Integer.parseInt(parameters[1]);
                remoteGridFTPPort = Integer.parseInt(parameters[2]);
                remoteLogging = Integer.parseInt(parameters[3]);
            } catch (NoHostsFoundException e) {
                // TODO Mostrare un messaggio in finestra "No hosts found"
                // e fermare l'esecuzione del programma (cercare stop)
                e.printStackTrace();
            }
        } else {
            remoteHost = host;
            remoteServicePort = Integer
                    .parseInt(Hosts.servicePorts[hostSelectedIndex]);
            remoteGridFTPPort = Integer
                    .parseInt(Hosts.gridFTPPorts[hostSelectedIndex]);
            remoteLogging = Integer
                    .parseInt(Hosts.loggingOptions[hostSelectedIndex]);
        }
    }

    protected static long getTrainTotalTimeStart() {
        return trainTotalTimeStart;
    }

    /**
     * Set the stop button
     */
    protected void setStopButton(JButton stopButton) {
        m_stopB = stopButton;
    }

    /**
     * Set the startAll button
     */
    protected void setStartAllButton(JButton startAllButton) {
        m_startAllB = startAllButton;
    }

    protected synchronized int getRunningNodes() {
        return runningNodes;
    }

    protected synchronized void lastOneCheck() {
        if (isLastOne()) {
            // condition was: m_log != null
            // && isLastOne()
            // && m_batchClassifierListeners.size() == 0

            long trainTimeElapsed = System.currentTimeMillis()
                    - getTrainTotalTimeStart();
            m_log.logMessage("");
            m_log.logMessage("Total time taken to to run all algorithms: "
                    + WUtil.showTime(trainTimeElapsed));
            m_log.logMessage("");
            m_log.logMessage("----------- @ -----------");
            m_log.logMessage("");
            m_log.statusMessage("Done.");
            m_startAllB.setEnabled(true);
            m_stopB.setEnabled(false);
            trainTotalTimeStart = 0;
        }
    }

}
