/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    CrossValidationFoldMaker.java
 *    Copyright (C) 2002 Mark Hall
 *
 */

/*
 *  This code has been partially modified and extended as part of Weka4WS.
 *  Modifications and extensions are distinctly marked.
 *  Copyright (C) 2007 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */

package weka4ws.gui.kflow;

import java.io.Serializable;
import java.util.Enumeration;
import java.util.Random;
import java.util.Vector;

import weka.core.Instances;
import weka.gui.beans.AbstractTrainAndTestSetProducer;
import weka.gui.beans.BeanCommon;
import weka.gui.beans.BeanVisual;
import weka.gui.beans.DataSetEvent;
import weka.gui.beans.DataSourceListener;
import weka.gui.beans.EventConstraints;
import weka.gui.beans.TestSetEvent;
import weka.gui.beans.TestSetListener;
import weka.gui.beans.TrainingSetEvent;
import weka.gui.beans.TrainingSetListener;
import weka.gui.beans.UserRequestAcceptor;
import weka4ws.core.Hosts;
import weka4ws.core.RemoteTask;
import weka4ws.stubs.service.TestOptions;

/**
 * Bean for splitting instances into training ant test sets according to a cross
 * validation
 * 
 * @author <a href="mailto:mhall@cs.waikato.ac.nz">Mark Hall</a>
 * @author <a href="mailto:mlackovic@deis.unical.it">Marco Lackovic</a>
 * @version $Revision: 1.7 $
 */
public class CrossValidationFoldMaker extends AbstractTrainAndTestSetProducer
        implements DataSourceListener, TrainingSetListener, TestSetListener,
        UserRequestAcceptor, EventConstraints, Serializable {

    /* ============== Weka4WS modifications ============== */

    private static final long serialVersionUID = 1L;

    /* =================================================== */

    private int m_numFolds = 10;

    private int m_randomSeed = 1;

    private Thread m_foldThread = null;

    public CrossValidationFoldMaker() {
        m_visual.loadIcons(BeanVisual.ICON_PATH
                + "CrossValidationFoldMaker.gif", BeanVisual.ICON_PATH
                + "CrossValidationFoldMaker_animated.gif");
        m_visual.setText("CrossValidationFoldMaker");
    }

    /**
     * Global info for this bean
     * 
     * @return a <code>String</code> value
     */
    public String globalInfo() {
        return "Split an incoming data set into cross validation folds. "
                + "Separate train and test sets are produced for each of the k folds.";
    }

    /**
     * Accept a training set
     * 
     * @param e
     *            a <code>TrainingSetEvent</code> value
     */
    public void acceptTrainingSet(TrainingSetEvent e) {
        Instances trainingSet = e.getTrainingSet();
        DataSetEvent dse = new DataSetEvent(this, trainingSet);
        acceptDataSet(dse);
    }

    /**
     * Accept a test set
     * 
     * @param e
     *            a <code>TestSetEvent</code> value
     */
    public void acceptTestSet(TestSetEvent e) {
        Instances testSet = e.getTestSet();
        DataSetEvent dse = new DataSetEvent(this, testSet);
        acceptDataSet(dse);
    }

    /* ============== Weka4WS modifications ============== */
    /**
     * Accept a data set
     * 
     * @param e
     *            a <code>DataSetEvent</code> value
     */
    public void acceptDataSet(final DataSetEvent e) {
        if (e.isStructureOnly()) {
            String filePath = ((DataSetEvent4WS) e).getFilePath();
            // Pass on structure to training and test set listeners
            TrainingSetEvent4WS tse = new TrainingSetEvent4WS(this, e
                    .getDataSet());
            tse.setFilePath(filePath);
            TestSetEvent4WS tsee = new TestSetEvent4WS(this, e.getDataSet());
            tsee.setFilePath(filePath);
            /* =================================================== */
            notifyTrainingSetProduced(tse);
            notifyTestSetProduced(tsee);
            return;
        }
        if (m_foldThread == null) {
            final Instances dataSet = new Instances(e.getDataSet());
            m_foldThread = new Thread() {
                public void run() {
                    try {
                        Random random = new Random(getSeed());
                        dataSet.randomize(random);
                        if (dataSet.classIndex() >= 0
                                && dataSet.attribute(dataSet.classIndex())
                                        .isNominal()) {
                            dataSet.stratify(getFolds());
                            if (m_logger != null) {
                                m_logger
                                        .logMessage("CrossValidationFoldMaker : "
                                                + "stratifying data");
                            }
                        }

                        for (int i = 0; i < getFolds(); i++) {
                            if (m_foldThread == null) {
                                if (m_logger != null) {
                                    m_logger
                                            .logMessage("Cross validation has been canceled!");
                                    m_logger.statusMessage("OK");
                                }
                                // exit gracefully
                                break;
                            }
                            Instances train = dataSet.trainCV(getFolds(), i,
                                    random);
                            Instances test = dataSet.testCV(getFolds(), i);

                            /* ============== Weka4WS modifications ============== */
                            String filePath = ((DataSetEvent4WS) e)
                                    .getFilePath();
                            TestOptions testOptions = new TestOptions();
                            testOptions
                                    .setTestMode(RemoteTask.CROSS_VALIDATION);
                            testOptions.setPercent(0);
                            testOptions.setNumFolds(getFolds());
                            // inform all training set listeners
                            TrainingSetEvent4WS tse = new TrainingSetEvent4WS(
                                    this, train);
                            tse.setFilePath(filePath);
                            tse.setSetNumber(i + 1);
                            tse.setMaxSetNumber(getFolds());
                            tse.setTestOptions(testOptions);
                            tse.setSeed(getSeed());
                            if (m_foldThread != null) {
                                // System.err.println("--Just before notify
                                // training set");
                                notifyTrainingSetProduced(tse);
                                // System.err.println("---Just after notify");
                            }

                            // inform all test set listeners
                            TestSetEvent4WS teste = new TestSetEvent4WS(this,
                                    test);
                            teste.setFilePath(filePath);
                            teste.setSetNumber(i + 1);
                            teste.setMaxSetNumber(getFolds());
                            teste.setTestOptions(testOptions);
                            teste.setSeed(getSeed());
                            /* =================================================== */
                            if (m_foldThread != null) {
                                notifyTestSetProduced(teste);
                            }
                        }
                    } catch (Exception ex) {
                        ex.printStackTrace();
                    } finally {
                        if (isInterrupted()) {
                            System.err.println("Cross validation interrupted");
                        }
                        block(false);
                    }
                }
            };
            m_foldThread.setPriority(Thread.MIN_PRIORITY);
            m_foldThread.start();

            // if (m_foldThread.isAlive()) {
            block(true);
            // }
            m_foldThread = null;
        }
    }

    /**
     * Notify all test set listeners of a TestSet event
     * 
     * @param tse
     *            a <code>TestSetEvent</code> value
     */
    private void notifyTestSetProduced(TestSetEvent4WS tse) {
        Vector l;
        synchronized (this) {
            l = (Vector) m_testListeners.clone();
        }
        if (l.size() > 0) {
            for (int i = 0; i < l.size(); i++) {
                // System.err.println("Notifying test listeners "
                // +"(cross validation fold maker)");

                TestSetListener tsLis = ((TestSetListener) l.elementAt(i));
                // TODO Questo controllo pu essere evitato in quanto effettuato
                // gi da TestSetNotifierThread e pu bloccare l'esecuzione concorrente
                if (tsLis instanceof ComputingNode
                        && ((ComputingNode) tsLis).getSelectedHostIndex() == Hosts.LOCAL_INDEX) {
                    ((TestSetListener) l.elementAt(i)).acceptTestSet(tse);
                } else if (tse.getSetNumber() == 1) {
                    TestSetNotifierThread notifyThread = new TestSetNotifierThread(
                            tsLis, tse);
                    notifyThread.start();
                }
            }
        }
    }

    /**
     * Notify all listeners of a TrainingSet event
     * 
     * @param tse
     *            a <code>TrainingSetEvent</code> value
     */
    protected void notifyTrainingSetProduced(TrainingSetEvent tse) {
        Vector l;
        synchronized (this) {
            l = (Vector) m_trainingListeners.clone();
        }

        if (l.size() > 0) {
            for (int i = 0; i < l.size(); i++) {
                // System.err.println("Notifying training listeners "
                // +"(cross validation fold maker)");
                ((TrainingSetListener) l.elementAt(i)).acceptTrainingSet(tse);
            }
        }
    }

    /**
     * Set the number of folds for the cross validation
     * 
     * @param numFolds
     *            an <code>int</code> value
     */
    public void setFolds(int numFolds) {
        m_numFolds = numFolds;
    }

    /**
     * Get the currently set number of folds
     * 
     * @return an <code>int</code> value
     */
    public int getFolds() {
        return m_numFolds;
    }

    /**
     * Tip text for this property
     * 
     * @return a <code>String</code> value
     */
    public String foldsTipText() {
        return "The number of train and test splits to produce";
    }

    /**
     * Set the seed
     * 
     * @param randomSeed
     *            an <code>int</code> value
     */
    public void setSeed(int randomSeed) {
        m_randomSeed = randomSeed;
    }

    /**
     * Get the currently set seed
     * 
     * @return an <code>int</code> value
     */
    public int getSeed() {
        return m_randomSeed;
    }

    /**
     * Tip text for this property
     * 
     * @return a <code>String</code> value
     */
    public String seedTipText() {
        return "The randomization seed";
    }

    /**
     * Stop any action
     */
    public void stop() {
        // tell the listenee (upstream bean) to stop
        if (m_listenee instanceof BeanCommon) {
            System.err.println("Listener is BeanCommon");
            ((BeanCommon) m_listenee).stop();
        }

        // stop the fold thread
        if (m_foldThread != null) {
            // m_buildThread.interrupt();
            Thread temp = m_foldThread;
            // m_buildThread.stop();
            m_foldThread = null;
            temp.interrupt();
        }
    }

    /**
     * Function used to stop code that calls acceptDataSet. This is needed as
     * cross validation is performed inside a separate thread of execution.
     * 
     * @param tf
     *            a <code>boolean</code> value
     */
    private synchronized void block(boolean tf) {
        if (tf) {
            try {
                // make sure the thread is still running before we block
                if (m_foldThread.isAlive()) {
                    wait();
                }
            } catch (InterruptedException ex) {
            }
        } else {
            notifyAll();
        }
    }

    /**
     * Return an enumeration of user requests
     * 
     * @return an <code>Enumeration</code> value
     */
    public Enumeration enumerateRequests() {
        Vector newVector = new Vector(0);
        if (m_foldThread != null) {
            newVector.addElement("Stop");
        }
        return newVector.elements();
    }

    /**
     * Perform the named request
     * 
     * @param request
     *            a <code>String</code> value
     * @exception IllegalArgumentException
     *                if an error occurs
     */
    public void performRequest(String request) {
        if (request.compareTo("Stop") == 0) {
            stop();
        } else {
            throw new IllegalArgumentException(request
                    + " not supported (CrossValidation)");
        }
    }

    /**
     * Returns true, if at the current time, the named event could be generated.
     * Assumes that the supplied event name is an event that could be generated
     * by this bean
     * 
     * @param eventName
     *            the name of the event in question
     * @return true if the named event could be generated at this point in time
     */
    public boolean eventGeneratable(String eventName) {
        if (m_listenee == null) {
            return false;
        }

        if (m_listenee instanceof EventConstraints) {
            if (((EventConstraints) m_listenee).eventGeneratable("dataSet")
                    || ((EventConstraints) m_listenee)
                            .eventGeneratable("trainingSet")
                    || ((EventConstraints) m_listenee)
                            .eventGeneratable("testSet")) {
                return true;
            } else {
                return false;
            }
        }
        return true;
    }
}
